import math
import os
from collections import defaultdict

import h5py
import numpy as np

from rabbit import common, h5pyutils

from wums import ioutils, logging, output_tools  # isort: skip

logger = logging.child_logger(__name__)


class TensorWriter:
    def __init__(
        self,
        sparse=False,
        systematic_type="log_normal",
        allow_negative_expectation=False,
        add_bin_by_bin_stat_to_data_cov=False,
    ):
        self.allow_negative_expectation = allow_negative_expectation

        self.systematic_type = systematic_type

        self.symmetric_tensor = True  # If all shape systematics are symmetrized the systematic tensor is symmetric leading to reduced memory and improved efficiency
        self.add_bin_by_bin_stat_to_data_cov = add_bin_by_bin_stat_to_data_cov  # add bin by bin statistical uncertainty to data covariance matrix

        self.signals = set()
        self.bkgs = set()

        self.channels = {}
        self.nbinschan = {}
        self.pseudodata_names = set()

        self.dict_systgroups = defaultdict(lambda: set())

        self.systsstandard = set()
        self.systsnoi = set()
        self.systsnoconstraint = set()
        self.systscovariance = set()

        self.sparse = sparse
        self.idxdtype = "int64"

        # temporary data
        self.dict_data_obs = {}  # [channel]
        self.dict_data_var = {}  # [channel]
        self.data_covariance = None
        self.dict_pseudodata = {}  # [channel][pseudodata]
        self.dict_norm = {}  # [channel][process]
        self.dict_sumw2 = {}  # [channel][process]
        self.dict_logkavg = {}  # [channel][proc][syst]
        self.dict_logkhalfdiff = {}  # [channel][proc][syst]
        self.dict_logkavg_indices = {}
        self.dict_logkhalfdiff_indices = {}

        self.clipSystVariations = False
        if self.clipSystVariations > 0.0:
            self.clip = np.abs(np.log(self.clipSystVariations))

        self.logkepsilon = math.log(
            1e-3
        )  # numerical cutoff in case of zeros in systematic variations

        # settings for writing out hdf5 files
        self.dtype = "float64"
        self.chunkSize = 4 * 1024**2

    def get_flat_values(self, h, flow=False):
        if hasattr(h, "values"):
            values = h.values(flow=flow)
        else:
            values = h
        return values.flatten().astype(self.dtype)

    def get_flat_variances(self, h, flow=False):
        if hasattr(h, "variances"):
            variances = h.variances(flow=flow)
        else:
            variances = h

        if (variances < 0.0).any():
            raise ValueError("Negative variances encountered")

        return variances.flatten().astype(self.dtype)

    def add_data(self, h, channel="ch0", variances=None):
        self._check_hist_and_channel(h, channel)
        if channel in self.dict_data_obs.keys():
            raise RuntimeError(f"Data histogram for channel '{channel}' already set.")
        self.dict_data_obs[channel] = self.get_flat_values(h)
        self.dict_data_var[channel] = self.get_flat_variances(
            h if variances is None else variances
        )

    def add_data_covariance(self, cov):
        self.data_covariance = cov if isinstance(cov, np.ndarray) else cov.values()

    def add_pseudodata(self, h, name=None, channel="ch0"):
        self._check_hist_and_channel(h, channel)
        if name is None:
            name = f"pseudodata_{len(self.pseudodata_names)}"
        self.pseudodata_names.add(name)
        if channel not in self.dict_pseudodata.keys():
            self.dict_pseudodata[channel] = {}
        if name in self.dict_pseudodata[channel].keys():
            raise RuntimeError(
                f"Pseudodata histogram '{name}' for channel '{channel}' already set."
            )
        self.dict_pseudodata[channel][name] = self.get_flat_values(h)

    def add_process(self, h, name, channel="ch0", signal=False, variances=None):
        self._check_hist_and_channel(h, channel)

        if name in self.dict_norm[channel].keys():
            raise RuntimeError(
                f"Nominal histogram for process '{name}' for channel '{channel}' already set."
            )

        if signal:
            self.signals.add(name)
        else:
            self.bkgs.add(name)

        self.dict_logkavg[channel][name] = {}
        self.dict_logkhalfdiff[channel][name] = {}
        if self.sparse:
            self.dict_logkavg_indices[channel][name] = {}
            self.dict_logkhalfdiff_indices[channel][name] = {}

        flow = self.channels[channel]["flow"]
        norm = self.get_flat_values(h, flow)
        sumw2 = self.get_flat_variances(h if variances is None else variances, flow)

        if not self.allow_negative_expectation:
            norm = np.maximum(norm, 0.0)
        if not np.all(np.isfinite(sumw2)):
            raise RuntimeError(
                f"{len(sumw2)-sum(np.isfinite(sumw2))} NaN or Inf values encountered in variances for {name}!"
            )
        if not np.all(np.isfinite(norm)):
            raise RuntimeError(
                f"{len(norm)-sum(np.isfinite(norm))} NaN or Inf values encountered in nominal histogram for {name}!"
            )

        self.dict_norm[channel][name] = norm
        self.dict_sumw2[channel][name] = sumw2

    def add_channel(self, axes, name=None, masked=False, flow=False):
        if flow and masked is False:
            raise NotImplementedError(
                "Keeping underflow/overflow is currently only supported for masked channels"
            )
        if name is None:
            name = f"ch{len(self.channels)}"
        logger.debug(f"Add new channel {name}")
        ibins = np.prod([a.extent if flow else a.size for a in axes])
        self.nbinschan[name] = ibins
        self.dict_norm[name] = {}
        self.dict_sumw2[name] = {}

        # add masked channels last and not masked channels first
        this_channel = {"axes": [a for a in axes], "masked": masked, "flow": flow}
        if masked:
            self.channels[name] = this_channel
        else:
            self.channels = {name: this_channel, **self.channels}

        self.dict_logkavg[name] = {}
        self.dict_logkhalfdiff[name] = {}
        if self.sparse:
            self.dict_logkavg_indices[name] = {}
            self.dict_logkhalfdiff_indices[name] = {}

    def _check_hist_and_channel(self, h, channel):

        if channel not in self.channels.keys():
            raise RuntimeError(f"Channel {channel} not known!")

        if hasattr(h, "axes"):
            axes = [a for a in h.axes]
            channel_axes = self.channels[channel]["axes"]

            if not all(np.allclose(a, axes[i]) for i, a in enumerate(channel_axes)):
                raise RuntimeError(
                    f"""
                    Histogram axes different have different edges from channel axes of channel {channel}
                    \nHistogram axes: {[a.edges for a in axes]}
                    \nChannel axes: {[a.edges for a in channel_axes]}
                    """
                )
        else:
            shape_in = h.shape
            shape_this = tuple([len(a) for a in self.channels[channel]["axes"]])
            if shape_in != shape_this:
                raise RuntimeError(
                    f"Shape of input object different from channel axes '{shape_in}' != '{shape_this}'"
                )

    def _compute_asym_syst(
        self,
        logkup,
        logkdown,
        name,
        process,
        channel,
        symmetrize="average",
        add_to_data_covariance=False,
        **kargs,
    ):
        var_name_out = name

        if symmetrize == "conservative":
            # symmetrize by largest magnitude of up and down variations
            logkavg_proc = np.where(
                np.abs(logkup) > np.abs(logkdown),
                logkup,
                logkdown,
            )
        elif symmetrize == "average":
            # symmetrize by average of up and down variations
            logkavg_proc = 0.5 * (logkup + logkdown)
        elif symmetrize in ["linear", "quadratic"]:
            # "linear" corresponds to a piecewise linear dependence of logk on theta
            # while "quadratic" corresponds to a quadratic dependence and leads
            # to a large variance
            diff_fact = np.sqrt(3.0) if symmetrize == "quadratic" else 1.0

            # split asymmetric variation into two symmetric variations
            logkavg_proc = 0.5 * (logkup + logkdown)
            logkdiffavg_proc = 0.5 * diff_fact * (logkup - logkdown)

            var_name_out = name + "SymAvg"
            var_name_out_diff = name + "SymDiff"

            # special case, book the extra systematic
            self.book_logk_avg(logkdiffavg_proc, channel, process, var_name_out_diff)
            self.book_systematic(
                var_name_out_diff,
                add_to_data_covariance=add_to_data_covariance,
                **kargs,
            )
        else:
            if add_to_data_covariance:
                raise RuntimeError(
                    "add_to_data_covariance requires symmetric uncertainties"
                )

            self.symmetric_tensor = False

            logkavg_proc = 0.5 * (logkup + logkdown)
            logkhalfdiff_proc = 0.5 * (logkup - logkdown)

            self.book_logk_halfdiff(logkhalfdiff_proc, channel, process, name)
        logkup = None
        logkdown = None

        return logkavg_proc, var_name_out

    def add_norm_systematic(
        self,
        name,
        process,
        channel,
        uncertainty,
        add_to_data_covariance=False,
        groups=None,
        symmetrize="average",
        **kargs,
    ):
        if not isinstance(process, (list, tuple, np.ndarray)):
            process = [process]

        if not isinstance(uncertainty, (list, tuple, np.ndarray)):
            uncertainty = [uncertainty]

        if len(uncertainty) != 1 and len(process) != len(uncertainty):
            raise RuntimeError(
                f"uncertainty must be either a scalar or list with the same length as the list of processes but len(process) = {len(process)} and len(uncertainty) = {len(uncertainty)}"
            )

        var_name_out = name

        systematic_type = "normal" if add_to_data_covariance else self.systematic_type

        for p, u in zip(process, uncertainty):
            norm = self.dict_norm[channel][p]
            if isinstance(u, (list, tuple, np.ndarray)):
                if len(u) != 2:
                    raise RuntimeError(
                        f"lnN uncertainty can only be a scalar for a symmetric or a list of 2 elements for asymmetric lnN uncertainties, but got a list of {len(u)} elements"
                    )
                # asymmetric lnN uncertainty
                syst_up = norm * u[0]
                syst_down = norm * u[1]

                logkup_proc = self.get_logk(
                    syst_up, norm, systematic_type=systematic_type
                )
                logkdown_proc = -self.get_logk(
                    syst_down, norm, systematic_type=systematic_type
                )

                logkavg_proc, var_name_out = self._compute_asym_syst(
                    logkup_proc,
                    logkdown_proc,
                    name,
                    process,
                    channel,
                    symmetrize=symmetrize,
                    add_to_data_covariance=add_to_data_covariance,
                    **kargs,
                )
            else:
                syst = norm * u
                logkavg_proc = self.get_logk(
                    syst, norm, systematic_type=systematic_type
                )

            self.book_logk_avg(logkavg_proc, channel, p, var_name_out)

        self.book_systematic(
            var_name_out,
            groups=groups,
            add_to_data_covariance=add_to_data_covariance,
            **kargs,
        )

    def add_systematic(
        self,
        h,
        name,
        process,
        channel,
        kfactor=1,
        mirror=True,
        symmetrize="average",
        add_to_data_covariance=False,
        **kargs,
    ):
        """
        h: either a single histogram with the systematic variation if mirror=True or a list of two histograms with the up and down variation
        """

        norm = self.dict_norm[channel][process]

        var_name_out = name

        systematic_type = "normal" if add_to_data_covariance else self.systematic_type

        flow = self.channels[channel]["flow"]

        if isinstance(h, (list, tuple)):
            self._check_hist_and_channel(h[0], channel)
            self._check_hist_and_channel(h[1], channel)

            syst_up = self.get_flat_values(h[0], flow=flow)
            syst_down = self.get_flat_values(h[1], flow=flow)

            logkup_proc = self.get_logk(
                syst_up, norm, kfactor, systematic_type=systematic_type
            )
            logkdown_proc = -self.get_logk(
                syst_down, norm, kfactor, systematic_type=systematic_type
            )

            logkavg_proc, var_name_out = self._compute_asym_syst(
                logkup_proc,
                logkdown_proc,
                name,
                process,
                channel,
                symmetrize,
                add_to_data_covariance,
                **kargs,
            )
        elif mirror:
            self._check_hist_and_channel(h, channel)
            syst = self.get_flat_values(h, flow=flow)
            logkavg_proc = self.get_logk(
                syst, norm, kfactor, systematic_type=systematic_type
            )
        else:
            raise RuntimeError(
                "Only one histogram given but mirror=False, can not construct a variation"
            )

        self.book_logk_avg(logkavg_proc, channel, process, var_name_out)
        self.book_systematic(
            var_name_out, add_to_data_covariance=add_to_data_covariance, **kargs
        )

    def get_logk(self, syst, norm, kfac=1.0, systematic_type=None):
        if not np.all(np.isfinite(syst)):
            raise RuntimeError(
                f"{len(syst)-sum(np.isfinite(syst))} NaN or Inf values encountered in systematic!"
            )

        # TODO clean this up and avoid duplication
        if systematic_type == "log_normal":
            # check if there is a sign flip between systematic and nominal
            _logk = kfac * np.log(syst / norm)
            _logk_view = np.where(
                np.equal(np.sign(norm * syst), 1),
                _logk,
                self.logkepsilon * np.ones_like(_logk),
            )

            # FIXME does this actually take effect since _logk_view is normally returned?
            if self.clipSystVariations > 0.0:
                _logk = np.clip(_logk, -self.clip, self.clip)

            return _logk_view
        elif systematic_type == "normal":
            _logk = kfac * (syst - norm)
            return _logk
        else:
            raise RuntimeError(
                f"Invalid systematic_type {systematic_type}, valid choices are 'log_normal' or 'normal'"
            )

    def book_logk_avg(self, *args):
        self.book_logk(
            self.dict_logkavg,
            self.dict_logkavg_indices,
            *args,
        )

    def book_logk_halfdiff(self, *args):
        self.book_logk(
            self.dict_logkhalfdiff,
            self.dict_logkhalfdiff_indices,
            *args,
        )

    def book_logk(
        self,
        dict_logk,
        dict_logk_indices,
        logk,
        channel,
        process,
        syst_name,
    ):
        norm = self.dict_norm[channel][process]
        # ensure that systematic tensor is sparse where normalization matrix is sparse
        logk = np.where(np.equal(norm, 0.0), 0.0, logk)
        if self.sparse:
            indices = np.transpose(np.nonzero(logk))
            dict_logk_indices[channel][process][syst_name] = indices
            dict_logk[channel][process][syst_name] = np.reshape(logk[indices], [-1])
        else:
            dict_logk[channel][process][syst_name] = logk

    def book_systematic(
        self,
        name,
        noi=False,
        constrained=True,
        add_to_data_covariance=False,
        groups=None,
    ):

        if add_to_data_covariance:
            if noi:
                raise ValueError(
                    f"{name} is maked as 'noi' but an 'noi' can't be added to the data covariance matrix."
                )
            self.systscovariance.add(name)
        elif not constrained:
            self.systsnoconstraint.add(name)
        else:
            self.systsstandard.add(name)

        if noi:
            self.systsnoi.add(name)

        # below only makes sense if this is an explicit nuisance parameter
        if not add_to_data_covariance:
            if groups is None:
                groups = [name]

            for group in groups:
                self.dict_systgroups[group].add(name)

    def write(self, outfolder="./", outfilename="rabbit_input.hdf5", args={}):

        if self.signals.intersection(self.bkgs):
            raise RuntimeError(
                f"Processes '{self.signals.intersection(self.bkgs)}' found as signal and background"
            )

        procs = sorted(list(self.signals)) + sorted(list(self.bkgs))
        nproc = len(procs)

        nbins = sum(
            [v for c, v in self.nbinschan.items() if not self.channels[c]["masked"]]
        )
        # nbinsfull including masked channels
        nbinsfull = sum([v for v in self.nbinschan.values()])

        logger.info(f"Write out nominal arrays")
        sumw = np.zeros([nbinsfull, nproc], self.dtype)
        sumw2 = np.zeros([nbinsfull, nproc], self.dtype)
        data_obs = np.zeros([nbins], self.dtype)
        data_var = np.zeros([nbins], self.dtype)
        pseudodata = np.zeros([nbins, len(self.pseudodata_names)], self.dtype)
        ibin = 0
        for chan, chan_info in self.channels.items():
            nbinschan = self.nbinschan[chan]

            for iproc, proc in enumerate(procs):
                if proc not in self.dict_norm[chan]:
                    continue

                sumw[ibin : ibin + nbinschan, iproc] = self.dict_norm[chan][proc]
                sumw2[ibin : ibin + nbinschan, iproc] = self.dict_sumw2[chan][proc]

            if not chan_info["masked"]:
                data_obs[ibin : ibin + nbinschan] = self.dict_data_obs[chan]
                data_var[ibin : ibin + nbinschan] = self.dict_data_var[chan]

                for idx, name in enumerate(self.pseudodata_names):
                    pseudodata[ibin : ibin + nbinschan, idx] = self.dict_pseudodata[
                        chan
                    ][name]

            ibin += nbinschan

        systs = self.get_systs()
        nsyst = len(systs)

        if self.symmetric_tensor:
            logger.info("No asymmetric systematics - write fully symmetric tensor")

        ibin = 0
        if self.sparse:
            logger.info(f"Write out sparse array")
            norm_sparse_size = 0
            norm_sparse_indices = np.zeros([norm_sparse_size, 2], self.idxdtype)
            norm_sparse_values = np.zeros([norm_sparse_size], self.dtype)

            logk_sparse_size = 0
            logk_sparse_normindices = np.zeros([logk_sparse_size, 1], self.idxdtype)
            logk_sparse_systindices = np.zeros([logk_sparse_size, 1], self.idxdtype)
            logk_sparse_values = np.zeros([logk_sparse_size], self.dtype)

            for chan in self.channels.keys():
                nbinschan = self.nbinschan[chan]
                dict_norm_chan = self.dict_norm[chan]
                dict_logkavg_chan_indices = self.dict_logkavg_indices[chan]
                dict_logkavg_chan_values = self.dict_logkavg[chan]

                for iproc, proc in enumerate(procs):
                    if proc not in dict_norm_chan:
                        continue
                    norm_proc = dict_norm_chan[proc]

                    norm_indices = np.transpose(np.nonzero(norm_proc))
                    norm_values = np.reshape(norm_proc[norm_indices], [-1])

                    nvals = len(norm_values)
                    oldlength = norm_sparse_size
                    norm_sparse_size = oldlength + nvals
                    norm_sparse_indices.resize([norm_sparse_size, 2])
                    norm_sparse_values.resize([norm_sparse_size])

                    out_indices = np.array([[ibin, iproc]]) + np.pad(
                        norm_indices, ((0, 0), (0, 1)), "constant"
                    )
                    norm_indices = None

                    norm_sparse_indices[oldlength:norm_sparse_size] = out_indices
                    out_indices = None

                    norm_sparse_values[oldlength:norm_sparse_size] = norm_values
                    norm_values = None

                    norm_idx_map = (
                        np.cumsum(np.not_equal(norm_proc, 0.0)) - 1 + oldlength
                    )

                    dict_logkavg_proc_indices = dict_logkavg_chan_indices[proc]
                    dict_logkavg_proc_values = dict_logkavg_chan_values[proc]

                    for isyst, syst in enumerate(systs):
                        if syst not in dict_logkavg_proc_indices.keys():
                            continue

                        logkavg_proc_indices = dict_logkavg_proc_indices[syst]
                        logkavg_proc_values = dict_logkavg_proc_values[syst]

                        nvals_proc = len(logkavg_proc_values)
                        oldlength = logk_sparse_size
                        logk_sparse_size = oldlength + nvals_proc
                        logk_sparse_normindices.resize([logk_sparse_size, 1])
                        logk_sparse_systindices.resize([logk_sparse_size, 1])
                        logk_sparse_values.resize([logk_sparse_size])

                        # first dimension of output indices are NOT in the dense [nbin,nproc] space, but rather refer to indices in the norm_sparse vectors
                        # second dimension is flattened in the [2,nsyst] space, where logkavg corresponds to [0,isyst] flattened to isyst
                        # two dimensions are kept in separate arrays for now to reduce the number of copies needed later
                        out_normindices = norm_idx_map[logkavg_proc_indices]
                        logkavg_proc_indices = None

                        logk_sparse_normindices[oldlength:logk_sparse_size] = (
                            out_normindices
                        )
                        logk_sparse_systindices[oldlength:logk_sparse_size] = isyst
                        out_normindices = None

                        logk_sparse_values[oldlength:logk_sparse_size] = (
                            logkavg_proc_values
                        )
                        logkavg_proc_values = None

                        if syst in self.dict_logkhalfdiff_indices[chan][proc].keys():
                            logkhalfdiff_proc_indices = self.dict_logkhalfdiff_indices[
                                chan
                            ][proc][syst]
                            logkhalfdiff_proc_values = self.dict_logkhalfdiff[chan][
                                proc
                            ][syst]

                            nvals_proc = len(logkhalfdiff_proc_values)
                            oldlength = logk_sparse_size
                            logk_sparse_size = oldlength + nvals_proc
                            logk_sparse_normindices.resize([logk_sparse_size, 1])
                            logk_sparse_systindices.resize([logk_sparse_size, 1])
                            logk_sparse_values.resize([logk_sparse_size])

                            # out_indices = np.array([[ibin,iproc,isyst,1]]) + np.pad(logkhalfdiff_proc_indices,((0,0),(0,3)),'constant')
                            # first dimension of output indices are NOT in the dense [nbin,nproc] space, but rather refer to indices in the norm_sparse vectors
                            # second dimension is flattened in the [2,nsyst] space, where logkhalfdiff corresponds to [1,isyst] flattened to nsyst + isyst
                            # two dimensions are kept in separate arrays for now to reduce the number of copies needed later
                            out_normindices = norm_idx_map[logkhalfdiff_proc_indices]
                            logkhalfdiff_proc_indices = None

                            logk_sparse_normindices[oldlength:logk_sparse_size] = (
                                out_normindices
                            )
                            logk_sparse_systindices[oldlength:logk_sparse_size] = (
                                nsyst + isyst
                            )
                            out_normindices = None

                            logk_sparse_values[oldlength:logk_sparse_size] = (
                                logkhalfdiff_proc_values
                            )
                            logkhalfdiff_proc_values = None

                    # free memory
                    dict_logkavg_proc_indices = None
                    dict_logkavg_proc_values = None

                # free memory
                norm_proc = None
                norm_idx_map = None

                ibin += nbinschan

            logger.info(f"Resize and sort sparse arrays into canonical order")
            # resize sparse arrays to actual length
            norm_sparse_indices.resize([norm_sparse_size, 2])
            norm_sparse_values.resize([norm_sparse_size])
            logk_sparse_normindices.resize([logk_sparse_size, 1])
            logk_sparse_systindices.resize([logk_sparse_size, 1])
            logk_sparse_values.resize([logk_sparse_size])

            # straightforward sorting of norm_sparse into canonical order
            norm_sparse_dense_shape = (nbinsfull, nproc)
            norm_sort_indices = np.argsort(
                np.ravel_multi_index(
                    np.transpose(norm_sparse_indices), norm_sparse_dense_shape
                )
            )
            norm_sparse_indices = norm_sparse_indices[norm_sort_indices]
            norm_sparse_values = norm_sparse_values[norm_sort_indices]

            # now permute the indices of the first dimension of logk_sparse corresponding to the resorting of norm_sparse

            # compute the inverse permutation from the sorting of norm_sparse
            # since the final indices are filled from here, need to ensure it has the correct data type
            logk_permute_indices = np.argsort(norm_sort_indices).astype(self.idxdtype)
            norm_sort_indices = None
            logk_sparse_normindices = logk_permute_indices[logk_sparse_normindices]
            logk_permute_indices = None
            logk_sparse_indices = np.concatenate(
                [logk_sparse_normindices, logk_sparse_systindices], axis=-1
            )

            # now straightforward sorting of logk_sparse into canonical order
            if self.symmetric_tensor:
                logk_sparse_dense_shape = (norm_sparse_indices.shape[0], nsyst)
            else:
                logk_sparse_dense_shape = (norm_sparse_indices.shape[0], 2 * nsyst)
            logk_sort_indices = np.argsort(
                np.ravel_multi_index(
                    np.transpose(logk_sparse_indices), logk_sparse_dense_shape
                )
            )
            logk_sparse_indices = logk_sparse_indices[logk_sort_indices]
            logk_sparse_values = logk_sparse_values[logk_sort_indices]
            logk_sort_indices = None

        else:
            logger.info(f"Write out dense array")
            # initialize with zeros, i.e. no variation
            norm = np.zeros([nbinsfull, nproc], self.dtype)
            if self.symmetric_tensor:
                logk = np.zeros([nbinsfull, nproc, nsyst], self.dtype)
            else:
                logk = np.zeros([nbinsfull, nproc, 2, nsyst], self.dtype)

            for chan in self.channels.keys():
                nbinschan = self.nbinschan[chan]
                dict_norm_chan = self.dict_norm[chan]

                for iproc, proc in enumerate(procs):
                    if proc not in dict_norm_chan:
                        continue

                    norm_proc = dict_norm_chan[proc]

                    norm[ibin : ibin + nbinschan, iproc] = norm_proc

                    dict_logkavg_proc = self.dict_logkavg[chan][proc]
                    dict_logkhalfdiff_proc = self.dict_logkhalfdiff[chan][proc]
                    for isyst, syst in enumerate(systs):
                        if syst not in dict_logkavg_proc.keys():
                            continue

                        if self.symmetric_tensor:
                            logk[ibin : ibin + nbinschan, iproc, isyst] = (
                                dict_logkavg_proc[syst]
                            )
                        else:
                            logk[ibin : ibin + nbinschan, iproc, 0, isyst] = (
                                dict_logkavg_proc[syst]
                            )
                            if syst in dict_logkhalfdiff_proc.keys():
                                logk[ibin : ibin + nbinschan, iproc, 1, isyst] = (
                                    dict_logkhalfdiff_proc[syst]
                                )

                ibin += nbinschan

        if self.data_covariance is None and (
            self.systscovariance or self.add_bin_by_bin_stat_to_data_cov
        ):
            # create data covariance
            self.data_covariance = np.diag(data_var)

        # write results to hdf5 file
        procSize = nproc * np.dtype(self.dtype).itemsize
        systSize = 2 * nsyst * np.dtype(self.dtype).itemsize
        amax = np.max([procSize, systSize])
        if amax > self.chunkSize:
            logger.info(
                f"Maximum chunk size in bytes was increased from {self.chunkSize} to {amax} to align with tensor sizes and allow more efficient reading/writing."
            )
            self.chunkSize = amax

        # create HDF5 file (chunk cache set to the chunk size since we can guarantee fully aligned writes
        if not os.path.isdir(outfolder):
            os.makedirs(outfolder)
        outpath = f"{outfolder}/{outfilename}"
        if len(outfilename.split(".")) < 2:
            outpath += ".hdf5"
        logger.info(f"Write output file {outpath}")
        f = h5py.File(outpath, rdcc_nbytes=self.chunkSize, mode="w")

        # propagate meta info into result file
        meta = {
            "meta_info": output_tools.make_meta_info_dict(
                args=args, wd=common.base_dir
            ),
            "channel_info": self.channels,
            "symmetric_tensor": self.symmetric_tensor,
            "systematic_type": self.systematic_type,
        }

        ioutils.pickle_dump_h5py("meta", meta, f)

        noiidxs = self.get_noiidxs()
        systsnoconstraint = self.get_systsnoconstraint()
        systgroups, systgroupidxs = self.get_systgroups()

        # save some lists of strings to the file for later use
        def create_dataset(
            name,
            content,
            length=None,
            dtype=h5py.special_dtype(vlen=str),
            compression="gzip",
        ):
            dimension = [len(content), length] if length else [len(content)]
            ds = f.create_dataset(
                f"h{name}", dimension, dtype=dtype, compression=compression
            )
            ds[...] = content

        create_dataset("procs", procs)
        create_dataset("signals", sorted(list(self.signals)))
        create_dataset("systs", systs)
        create_dataset("systsnoconstraint", systsnoconstraint)
        create_dataset("systgroups", systgroups)
        create_dataset(
            "systgroupidxs",
            systgroupidxs,
            dtype=h5py.special_dtype(vlen=np.dtype("int32")),
        )
        create_dataset("noiidxs", noiidxs, dtype="int32")
        create_dataset("pseudodatanames", [n for n in self.pseudodata_names])

        # create h5py datasets with optimized chunk shapes
        nbytes = 0

        constraintweights = self.get_constraintweights(self.dtype)
        nbytes += h5pyutils.writeFlatInChunks(
            constraintweights, f, "hconstraintweights", maxChunkBytes=self.chunkSize
        )
        constraintweights = None

        nbytes += h5pyutils.writeFlatInChunks(
            data_obs, f, "hdata_obs", maxChunkBytes=self.chunkSize
        )
        if np.any(data_var != data_obs):
            nbytes += h5pyutils.writeFlatInChunks(
                data_var, f, "hdata_var", maxChunkBytes=self.chunkSize
            )
            data_var = None
        data_obs = None

        nbytes += h5pyutils.writeFlatInChunks(
            pseudodata, f, "hpseudodata", maxChunkBytes=self.chunkSize
        )
        pseudodata = None

        if self.data_covariance is not None:
            for syst in self.systscovariance:
                systv = np.zeros(shape=(nbinsfull, 1), dtype=self.dtype)

                ibin = 0
                for chan in self.channels.keys():
                    nbinschan = self.nbinschan[chan]
                    dict_norm_chan = self.dict_norm[chan]

                    for proc in procs:
                        if proc not in dict_norm_chan:
                            continue

                        dict_logkavg_proc = self.dict_logkavg[chan][proc]

                        if syst not in dict_logkavg_proc.keys():
                            continue

                        systv[ibin : ibin + nbinschan, 0] += dict_logkavg_proc[syst]

                    ibin += nbinschan

                self.data_covariance[...] += systv @ systv.T

            full_cov = (
                np.add(self.data_covariance, np.diag(sumw2))
                if self.add_bin_by_bin_stat_to_data_cov
                else self.data_covariance
            )
            full_cov_inv = np.linalg.inv(full_cov)

            nbytes += h5pyutils.writeFlatInChunks(
                full_cov_inv,
                f,
                "hdata_cov_inv",
                maxChunkBytes=self.chunkSize,
            )

        nbytes += h5pyutils.writeFlatInChunks(
            sumw, f, "hsumw", maxChunkBytes=self.chunkSize
        )

        nbytes += h5pyutils.writeFlatInChunks(
            sumw2, f, "hsumw2", maxChunkBytes=self.chunkSize
        )

        if self.sparse:
            nbytes += h5pyutils.writeSparse(
                norm_sparse_indices,
                norm_sparse_values,
                norm_sparse_dense_shape,
                f,
                "hnorm_sparse",
                maxChunkBytes=self.chunkSize,
            )
            norm_sparse_indices = None
            norm_sparse_values = None
            nbytes += h5pyutils.writeSparse(
                logk_sparse_indices,
                logk_sparse_values,
                logk_sparse_dense_shape,
                f,
                "hlogk_sparse",
                maxChunkBytes=self.chunkSize,
            )
            logk_sparse_indices = None
            logk_sparse_values = None
        else:
            nbytes += h5pyutils.writeFlatInChunks(
                norm, f, "hnorm", maxChunkBytes=self.chunkSize
            )
            norm = None
            nbytes += h5pyutils.writeFlatInChunks(
                logk, f, "hlogk", maxChunkBytes=self.chunkSize
            )
            logk = None

        logger.info(f"Total raw bytes in arrays = {nbytes}")

    def get_systsstandard(self):
        return list(common.natural_sort(self.systsstandard))

    def get_systsnoi(self):
        return list(common.natural_sort(self.systsnoi))

    def get_systsnoconstraint(self):
        return list(common.natural_sort(self.systsnoconstraint))

    def get_systs(self):
        return self.get_systsnoconstraint() + self.get_systsstandard()

    def get_constraintweights(self, dtype):
        systs = self.get_systs()
        constraintweights = np.ones([len(systs)], dtype=dtype)
        for syst in self.get_systsnoconstraint():
            constraintweights[systs.index(syst)] = 0.0
        return constraintweights

    def get_groups(self, group_dict):
        systs = self.get_systs()
        groups = []
        idxs = []
        for group, members in common.natural_sort_dict(group_dict).items():
            groups.append(group)
            idx = []
            for syst in members:
                idx.append(systs.index(syst))
            idxs.append(idx)
        return groups, idxs

    def get_noiidxs(self):
        # list of indeces of nois w.r.t. systs
        systs = self.get_systs()
        idxs = []
        for noi in self.get_systsnoi():
            idxs.append(systs.index(noi))
        return idxs

    def get_systgroups(self):
        # list of groups of systematics (nuisances) and lists of indexes
        return self.get_groups(self.dict_systgroups)



#!/usr/bin/env python3

import argparse
import itertools
import os

import hist
import matplotlib.pyplot as plt
import mplhep as hep
import numpy as np
from wums import boostHistHelpers as hh
from wums import logging, output_tools, plot_tools

from rabbit import debugdata, inputdata


def parseArgs():

    # choices for legend padding
    choices_padding = ["auto", "lower left", "lower right", "upper left", "upper right"]

    parser = argparse.ArgumentParser()

    parser.add_argument("infile", help="Output h5py input data file")
    parser.add_argument(
        "--logx", action="store_true", help="Enable log scale for x axis"
    )
    parser.add_argument(
        "--logy", action="store_true", help="Enable log scale for y axis"
    )
    parser.add_argument(
        "--invertAxes",
        action="store_true",
        help="Invert the order of the axes when plotting",
    )
    parser.add_argument("--noData", action="store_true", help="Don't plot data")
    parser.add_argument("--noRatio", action="store_true", help="Don't plot the ratio")
    parser.add_argument(
        "--noStack", action="store_true", help="Don't plot the individual processes"
    )
    parser.add_argument(
        "--processes", type=str, nargs="*", default=[], help="Select processes"
    )
    parser.add_argument(
        "--channels", type=str, nargs="*", default=None, help="Select channels"
    )
    parser.add_argument(
        "--splitByProcess",
        action="store_true",
        help="Make a separate plot for each of the selected processes",
    )
    parser.add_argument(
        "--selectionAxes",
        type=str,
        nargs="*",
        default=["charge", "passIso", "passMT"],
        help="List of axes where for each bin a seperate plot is created",
    )
    parser.add_argument(
        "--select",
        type=int,
        nargs="*",
        default=[],
        help="Select specific bins of the selectionAxis e.g. '0 1' to select the first bin of the first axis and second bin of the second axis",
    )
    parser.add_argument(
        "--hists",
        type=str,
        nargs="*",
        default=None,
        help="List of hists to plot; dash separated for unrolled hists",
    )
    parser.add_argument(
        "--normToData", action="store_true", help="Normalize MC to data", default=False
    )
    parser.add_argument(
        "--pseudodata",
        type=str,
        default=None,
        help="Pseuododata name to plot instead of data.",
    )
    parser.add_argument(
        "--dataName", type=str, default="Data", help="Data name for plot labeling"
    )
    parser.add_argument(
        "--xlabel", type=str, default=None, help="x-axis label for plot labeling"
    )
    parser.add_argument(
        "--ylabel", type=str, default=None, help="y-axis label for plot labeling"
    )
    parser.add_argument(
        "--processGrouping", type=str, default=None, help="key for grouping processes"
    )
    parser.add_argument(
        "--extraTextLoc",
        type=float,
        nargs="*",
        default=None,
        help="Location in (x,y) for additional text, aligned to upper left",
    )
    # variations
    parser.add_argument(
        "--varName", type=str, nargs="*", default=[], help="Name of variation hist"
    )
    parser.add_argument(
        "--varLabel",
        type=str,
        nargs="*",
        default=[],
        help="Label(s) of variation hist for plotting",
    )
    parser.add_argument(
        "--varColor", type=str, nargs="*", default=[], help="Variation colors"
    )
    parser.add_argument(
        "--oneSidedVariations",
        action="store_true",
        help="Plot variations one sided",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        type=int,
        default=3,
        choices=[0, 1, 2, 3, 4],
        help="Set verbosity level with logging, the larger the more verbose",
    )
    parser.add_argument(
        "--noColorLogger", action="store_true", help="Do not use logging with colors"
    )
    parser.add_argument(
        "-o",
        "--outpath",
        type=str,
        default=os.path.expanduser("./test"),
        help="Base path for output",
    )
    parser.add_argument(
        "--config",
        type=str,
        default=None,
        help="Path to config file for style formatting",
    )
    parser.add_argument(
        "--eoscp",
        action="store_true",
        help="Override use of xrdcp and use the mount instead",
    )
    parser.add_argument(
        "-p", "--postfix", type=str, help="Postfix for output file name"
    )
    parser.add_argument(
        "--lumi",
        type=float,
        default=16.8,
        help="Luminosity used in the fit, needed to get the absolute cross section",
    )
    parser.add_argument(
        "--title",
        default="Rabbit",
        type=str,
        help="Title to be printed in upper left",
    )
    parser.add_argument(
        "--subtitle",
        default="",
        type=str,
        help="Subtitle to be printed after title",
    )
    parser.add_argument("--titlePos", type=int, default=2, help="title position")
    parser.add_argument(
        "--legPos", type=str, default="upper right", help="Set legend position"
    )
    parser.add_argument(
        "--legSize",
        type=str,
        default="small",
        help="Legend text size (small: axis ticks size, large: axis label size, number)",
    )
    parser.add_argument(
        "--legCols", type=int, default=2, help="Number of columns in legend"
    )
    parser.add_argument(
        "--legPadding",
        type=str,
        default="auto",
        choices=choices_padding,
        help="Where to put empty entries in legend",
    )
    parser.add_argument(
        "--lowerLegPos",
        type=str,
        default="upper left",
        help="Set lower legend position",
    )
    parser.add_argument(
        "--lowerLegCols", type=int, default=2, help="Number of columns in lower legend"
    )
    parser.add_argument(
        "--lowerLegPadding",
        type=str,
        default="auto",
        choices=choices_padding,
        help="Where to put empty entries in lower legend",
    )
    parser.add_argument(
        "--noSciy",
        action="store_true",
        help="Don't allow scientific notation for y axis",
    )
    parser.add_argument(
        "--yscale",
        type=float,
        help="Scale the upper y axis by this factor (useful when auto scaling cuts off legend)",
    )
    parser.add_argument(
        "--ylim",
        type=float,
        nargs=2,
        help="Min and max values for y axis (if not specified, range set automatically)",
    )
    parser.add_argument("--xlim", type=float, nargs=2, help="min and max for x axis")
    parser.add_argument(
        "--rrange",
        type=float,
        nargs=2,
        default=[0.9, 1.1],
        help="y range for ratio plot",
    )
    parser.add_argument(
        "--scaleTextSize",
        type=float,
        default=1.0,
        help="Scale all text sizes by this number",
    )
    parser.add_argument(
        "--customFigureWidth",
        type=float,
        default=None,
        help="Use a custom figure width, otherwise chosen automatic",
    )

    args = parser.parse_args()

    return args


def make_plots(
    args,
    config,
    outdir,
    indata,
    hists_proc,
    hist_data,
    *opts,
    labels=None,
    colors=None,
    procs=None,
    **info,
):
    # make full unrollsed plot and lower dimensional projections

    all_axes_names = [n for n in hists_proc[0].axes.name]
    if args.invertAxes:
        logger.info("invert axes order")
        all_axes_names = all_axes_names[::-1]

    axes_combinations = all_axes_names[:]
    # make lower dimensional combinations of axes
    for n in range(2, len(all_axes_names) + 1):
        axes_combinations += [k for k in itertools.combinations(all_axes_names, n)]

    for axes_names in axes_combinations:
        if isinstance(axes_names, str):
            axes_names = [axes_names]

        if args.hists:
            if not any(set(axes_names) == set(h.split("-")) for h in args.hists):
                continue

        logger.info(f"Make plot(s) with axes {axes_names}")

        if args.splitByProcess:

            for i, hp in enumerate(hists_proc):
                if (hp.values() != 0).sum() == 0:
                    # skip processes with all 0
                    continue

                make_plot(
                    args,
                    config,
                    outdir,
                    indata,
                    [hp],
                    hist_data,
                    axes_names=axes_names,
                    *opts,
                    labels=labels[i],
                    colors=colors[i],
                    procs=procs[i],
                    **info,
                )
        else:
            make_plot(
                args,
                config,
                outdir,
                indata,
                hists_proc,
                hist_data,
                axes_names=axes_names,
                *opts,
                labels=labels,
                colors=colors,
                procs=procs,
                **info,
            )


def make_plot(
    args,
    config,
    outdir,
    indata,
    hists_proc,
    hist_data,
    hists_syst_up,
    hists_syst_dn,
    axes_names,
    selections=None,
    selection_edges=None,
    channel="",
    colors=[],
    labels=[],
    procs=[],
    rlabel="1/Pred.",
    density=False,
    systematics=None,
    colors_syst=None,
    labels_syst=None,
):
    if args.processGrouping is not None:
        hists_proc, labels, colors, procs = config.process_grouping(
            args.processGrouping, hists_proc, procs
        )

    if any(x in axes_names for x in ["ptll", "mll", "ptVgen", "ptVGen"]):
        # in case of variable bin width normalize to unit
        binwnorm = 1.0
        ylabel = "Events/unit"
    else:
        binwnorm = None
        ylabel = "Events/bin"

    if args.ylabel is not None:
        ylabel = args.ylabel

    if len(axes_names) == 1:
        fu = lambda h: h.project(*axes_names)
    else:
        fu = lambda h: hh.unrolledHist(h, binwnorm=binwnorm, obs=axes_names)

    # make 1D histograms
    h_stack = [fu(h) for h in hists_proc]

    if hist_data is not None:
        h_data = fu(hist_data)

    xlabel = plot_tools.get_axis_label(config, axes_names, args.xlabel)

    if args.splitByProcess:
        hists_pred = h_stack
    else:
        hists_pred = [hh.sumHists(h_stack)]

        if args.normToData:
            scale = h_data.values().sum() / hists_pred[0].values().sum()
            h_stack = [hh.scaleHist(h, scale) for h in h_stack]
            hists_pred[0] = hh.scaleHist(hists_pred[0], scale)
            hists_syst_up = [hh.scaleHist(h, scale) for h in hists_syst_up]
            hists_syst_dn = [hh.scaleHist(h, scale) for h in hists_syst_dn]

    # loop over all processes if plots for each process is requested, or inclusive otherwise
    for i, h_pred in enumerate(hists_pred):

        infos_figure = dict(
            xlabel=xlabel,
            ylabel=ylabel,
            logy=args.logy,
            logx=args.logx,
            xlim=args.xlim,
            ylim=args.ylim,
        )
        if not args.noRatio:
            fig, ax1, ratio_axes = plot_tools.figureWithRatio(
                h_pred, rlabel=rlabel, rrange=args.rrange, **infos_figure
            )
            ax2 = ratio_axes[-1]
        else:
            fig, ax1 = plot_tools.figure(h_pred, **infos_figure)

        if args.noStack or args.splitByProcess:
            hep.histplot(
                h_pred,
                xerr=False,
                yerr=False,
                histtype="step",
                color="black",
                label=labels[i] if args.splitByProcess else "Prediction",
                binwnorm=binwnorm,
                ax=ax1,
                zorder=1,
                flow="none",
            )
        else:
            hep.histplot(
                h_stack,
                xerr=False,
                yerr=False,
                histtype="fill",
                color=colors,
                label=labels,
                stack=True,
                density=False,
                binwnorm=binwnorm,
                ax=ax1,
                zorder=1,
                flow="none",
            )

        if hist_data is not None:
            hep.histplot(
                h_data,
                yerr=True,
                histtype="errorbar",
                color="black",
                label=args.dataName,
                binwnorm=binwnorm,
                ax=ax1,
                alpha=1.0,
                zorder=2,
                flow="none",
            )

        for hup, hdn, color, label in zip(
            hists_syst_up, hists_syst_dn, colors_syst, labels_syst
        ):
            if args.splitByProcess:
                hup = hup[{"processes": procs}]
                hdn = hdn[{"processes": procs}]
            else:
                hup = hup[{"processes": hist.sum}]
                hdn = hdn[{"processes": hist.sum}]

            hup = fu(hup)
            hdn = fu(hdn)

            hep.histplot(
                hup,
                xerr=False,
                yerr=False,
                histtype="step",
                color=color,
                label=label,
                binwnorm=binwnorm,
                ax=ax1,
                zorder=1,
                flow="none",
            )
            if not args.oneSidedVariations:
                hep.histplot(
                    hdn,
                    xerr=False,
                    yerr=False,
                    histtype="step",
                    color=color,
                    linestyle="--",
                    binwnorm=binwnorm,
                    ax=ax1,
                    zorder=1,
                    flow="none",
                )

            if not args.noRatio:
                h_ratios = (
                    [hh.divideHists(hup, h_pred)]
                    if args.oneSidedVariations
                    else [hh.divideHists(hup, h_pred), hh.divideHists(hdn, h_pred)]
                )
                hep.histplot(
                    h_ratios,
                    xerr=False,
                    yerr=False,
                    histtype="step",
                    color=color,
                    linestyle=["-", "--"][: len(h_ratios)],
                    ax=ax2,
                    linewidth=2,
                    flow="none",
                )

        if not args.noRatio:
            hep.histplot(
                hh.divideHists(
                    h_pred,
                    h_pred,
                    cutoff=1e-8,
                    rel_unc=True,
                    flow=False,
                    by_ax_name=False,
                ),
                histtype="step",
                color="black",
                alpha=0.5,
                yerr=False,
                ax=ax2,
                linewidth=1,
                flow="none",
            )

            if hist_data is not None:
                hep.histplot(
                    hh.divideHists(h_data, h_pred, cutoff=0.01, rel_unc=True),
                    histtype="errorbar",
                    color="black",
                    yerr=True,
                    linewidth=2,
                    ax=ax2,
                )

        scale = max(1, np.divide(*ax1.get_figure().get_size_inches()) * 0.3)

        if selections is not None:
            text = []
            for i, (key, idx) in enumerate(selections.items()):
                lo, hi = selection_edges[i]
                if key == "charge":
                    label = f"charge = {'-1' if hi==0 else '+1'}"
                else:
                    label = plot_tools.get_axis_label(config, key, with_unit=False)
                    if lo != None:
                        label = f"{lo} < {label}"
                    if hi != None:
                        label = f"{label} < {hi}"
                text.append(label)

            text_loc = (
                args.extraTextLoc if args.extraTextLoc else [0.05, 0.96 - i * 0.08]
            )
            plot_tools.wrap_text(
                text,
                ax1,
                # 0.05,
                *text_loc,
                # 0.96 - i * 0.08,
                ha="left",
                text_size="small",
            )

        if not args.noRatio:
            plot_tools.fix_axes(ax1, ax2, fig, yscale=args.yscale, noSci=args.noSciy)
        else:
            plot_tools.fix_axes(ax1, yscale=args.yscale, logy=args.logy)

        plot_tools.add_decor(
            ax1,
            args.title,
            args.subtitle,
            data=hist_data is not None or "Nonprompt" in labels,
            lumi=args.lumi,  # if args.dataName == "Data" and not args.noData else None,
            loc=args.titlePos,
            text_size=args.legSize,
        )

        plot_tools.addLegend(
            ax1, ncols=args.legCols, loc=args.legPos, text_size=args.legSize
        )

        outfile = "hist_"
        if not args.noStack:
            outfile += "stack_"
        if args.splitByProcess:
            outfile += f"{procs}_"
        outfile += "_".join(axes_names)
        outfile += f"_{channel}"
        if selections is not None:
            outfile += "_" + "_".join([f"{a}{i}" for a, i in selections.items()])
        if args.postfix:
            outfile += f"_{args.postfix}"
        plot_tools.save_pdf_and_png(outdir, outfile)

        output_tools.write_index_and_log(
            outdir,
            outfile,
            analysis_meta_info={"setupRabbit": indata.metadata["meta_info"]},
            args=args,
        )


def main():
    args = parseArgs()
    global logger
    logger = logging.setup_logger(__file__, args.verbose, args.noColorLogger)

    plt.rcParams["font.size"] = plt.rcParams["font.size"] * args.scaleTextSize

    indata = inputdata.FitInputData(args.infile, pseudodata=args.pseudodata)

    debug = debugdata.FitDebugData(indata)

    config = plot_tools.load_config(args.config)

    syst_labels = getattr(config, "systematics_labels", {})
    systematics = []
    colors_syst = []
    labels_syst = []
    for syst, color, label in itertools.zip_longest(
        args.varName, args.varColor, args.varLabel
    ):
        if syst not in indata.systs.astype(str):
            logger.error(f"Syst {syst} not available, skip!")
            continue
        systematics.append(syst)
        colors_syst.append(color if color is not None else "black")
        labels_syst.append(label if label is not None else syst_labels.get(syst, syst))

    if args.oneSidedVariations:
        labels_syst = [l.replace(r"\pm", "+") for l in labels_syst]

    outdir = output_tools.make_plot_dir(args.outpath, eoscp=args.eoscp)

    for channel in indata.channel_info.keys():
        if args.channels is not None and channel not in args.channels:
            continue
        logger.info(f"Make plots for channel: {channel}")

        hist_proc = debug.nominal_hists[channel]
        procs = [p for p in hist_proc.axes["processes"]]

        if len(args.processes):
            procs_tmp = procs[:]
            procs = []
            for p in args.processes:
                if p not in procs_tmp:
                    logger.warning(f"Process {p} requested but not found, skip")
                    continue
                procs.append(p)

        if hasattr(config, "get_labels_colors_procs_sorted"):
            labels, colors, procs = config.get_labels_colors_procs_sorted(procs)
        else:
            labels = procs[:]
            cmap = plt.get_cmap("tab10")
            proc_colors = getattr(config, "process_colors", {})
            colors = [proc_colors.get(p, cmap(i % cmap.N)) for i, p in enumerate(procs)]

        hists_proc = [hist_proc[{"processes": p}] for p in procs]

        if len(systematics):
            hist_syst = debug.syst_hists[channel]
            hists_syst_dn = [
                hist_syst[{"DownUp": "Down", "systs": s}] for s in systematics
            ]
            hists_syst_up = [
                hist_syst[{"DownUp": "Up", "systs": s}] for s in systematics
            ]
        else:
            hists_syst_dn = []
            hists_syst_up = []

        # setup data histogram
        if (
            args.noData
            or args.splitByProcess
            or channel not in debug.data_obs_hists.keys()
        ):
            hist_data = None
        else:
            hist_data_tmp = debug.data_obs_hists[channel]

            # poisson errors on data hist for correct errors in ratio plot
            hist_data = hist.Hist(*hist_data_tmp.axes, storage=hist.storage.Weight())
            hist_data.values(flow=True)[...] = hist_data_tmp.values(flow=True)
            hist_data.variances(flow=True)[...] = hist_data_tmp.variances(flow=True)

        info = dict(
            channel=channel,
            systematics=systematics,
            colors_syst=colors_syst,
            labels_syst=labels_syst,
        )

        # make plots in slices (e.g. for charge plus an minus separately)
        selection_axes = [
            hists_proc[0].axes[n]
            for n in args.selectionAxes
            if n in hists_proc[0].axes.name
        ]
        if len(selection_axes) > 0:
            other_axes = [a for a in hists_proc[0].axes if a not in selection_axes]
            if len(args.select):
                bin_combinations = [args.select]
            else:
                selection_bins = [
                    np.arange(a.size)
                    for a in hists_proc[0].axes
                    if a.name in args.selectionAxes
                ]
                bin_combinations = itertools.product(*selection_bins)

            for bins in bin_combinations:
                idxs = {a.name: i for a, i in zip(selection_axes, bins)}
                selection_edges = [
                    (a.edges[i], a.edges[i + 1] if len(a.edges - 1) > i else None)
                    for a, i in zip(selection_axes, bins)
                ]

                hs_proc = [h[idxs] for h in hists_proc]

                if hist_data is not None:
                    h_data = hist_data[idxs]
                else:
                    h_data = None

                hs_syst_dn = [h[idxs] for h in hists_syst_dn]
                hs_syst_up = [h[idxs] for h in hists_syst_up]

                make_plots(
                    args,
                    config,
                    outdir,
                    indata,
                    hs_proc,
                    h_data,
                    hs_syst_dn,
                    hs_syst_up,
                    selections=idxs,
                    selection_edges=selection_edges,
                    labels=labels,
                    colors=colors,
                    procs=procs,
                    **info,
                )
        else:
            make_plots(
                args,
                config,
                outdir,
                indata,
                hists_proc,
                hist_data,
                hists_syst_dn,
                hists_syst_up,
                labels=labels,
                colors=colors,
                procs=procs,
                **info,
            )

    if output_tools.is_eosuser_path(args.outpath) and args.eoscp:
        output_tools.copy_to_eos(outdir, args.outpath)


if __name__ == "__main__":
    main()



import os

from wums import logging

logger = logging.child_logger(__name__)


class DatacardParser:
    """
    Parser for Combine Datacard format files
    """

    def __init__(self):
        self.imax = None  # number of channels
        self.jmax = None  # number of backgrounds
        self.kmax = None  # number of nuisance parameters
        self.bins = []  # list of bin/channel names
        self.observations = {}  # observations for each bin
        self.processes = []  # list of process names
        self.process_indices = {}  # numerical indices for processes
        self.rates = {}  # expected rates for each bin and process
        self.systematics = []  # list of systematic uncertainties
        self.shapes = []  # shape directives
        self.bin_process_map = {}  # mapping of bins to processes
        self.param_lines = []  # param directives
        self.rate_params = []  # rate param directives
        self.group_lines = []  # group directives
        self.nuisance_edits = []  # nuisance edit directives
        self.max_depth = (
            0  # how many directories are there to find the histograms in the root files
        )

    def parse_file(self, filename):
        """Parse a datacard file"""
        if not os.path.exists(filename):
            raise FileNotFoundError(f"Datacard file not found: {filename}")

        with open(filename, "r") as file:
            content = file.read()

        # Remove comments (lines starting with #)
        lines = [line.strip() for line in content.split("\n")]
        lines = [line for line in lines if line and not line.startswith("#")]

        # Parse header section (imax, jmax, kmax)
        self._parse_header(lines)

        # Parse bin and observation section
        self._parse_observations(lines)

        # Parse process and rate section
        self._parse_processes_and_rates(lines)

        # Parse shapes section
        self._parse_shapes(lines, filename)

        # Parse systematics section
        self._parse_systematics(lines)

        # Parse additional directives
        self._parse_additional_directives(lines)

        return self

    def _parse_header(self, lines):
        """Parse the header section with imax, jmax, kmax"""
        for line in lines:
            if line.startswith("imax"):
                parts = line.split()
                self.imax = parts[1] if parts[1] != "*" else None
            elif line.startswith("jmax"):
                parts = line.split()
                self.jmax = parts[1] if parts[1] != "*" else None
            elif line.startswith("kmax"):
                parts = line.split()
                self.kmax = parts[1] if parts[1] != "*" else None

    def _parse_observations(self, lines):
        """Parse the bin and observation section"""
        bin_line = None
        obs_line = None

        for i, line in enumerate(lines):
            if (
                line.startswith("bin")
                and not bin_line
                and not self._is_process_bin_line(lines, i)
            ):
                bin_line = line
            elif line.startswith("observation") and not obs_line:
                obs_line = line

        if bin_line and obs_line:
            bin_parts = bin_line.split()[1:]
            obs_parts = obs_line.split()[1:]

            self.bins = bin_parts

            if len(bin_parts) == len(obs_parts):
                for i, bin_name in enumerate(bin_parts):
                    self.observations[bin_name] = float(obs_parts[i])

    def _is_process_bin_line(self, lines, index):
        """Check if a bin line is part of the process section"""
        if index > 0 and index < len(lines) - 2:
            next_line = lines[index + 1]
            if next_line.startswith("process"):
                return True
        return False

    def _parse_processes_and_rates(self, lines):
        """Parse the process and rate section"""
        bin_line = None
        process_name_line = None
        process_index_line = None
        rate_line = None

        # Find the three consecutive lines
        for i in range(len(lines) - 3):
            if (
                lines[i].startswith("bin")
                and lines[i + 1].startswith("process")
                and lines[i + 2].startswith("process")
                and lines[i + 3].startswith("rate")
            ):
                bin_line = lines[i]
                process_name_line = lines[i + 1]
                process_index_line = lines[i + 2]
                rate_line = lines[i + 3]
                break

        if bin_line and process_name_line and process_index_line and rate_line:
            bins = bin_line.split()[1:]
            process_names = process_name_line.split()[1:]
            process_indices = process_index_line.split()[1:]
            rates = rate_line.split()[1:]

            self.processes = list(
                dict.fromkeys(process_names)
            )  # Remove duplicates while preserving order

            # Map process names to their indices
            for name, idx in zip(process_names, process_indices):
                self.process_indices[name] = int(idx)

            # Map bins to processes with rates
            for i in range(len(bins)):
                bin_name = bins[i]
                process_name = process_names[i]
                rate = float(rates[i])

                if bin_name not in self.bin_process_map:
                    self.bin_process_map[bin_name] = {}

                self.bin_process_map[bin_name][process_name] = rate

                key = (bin_name, process_name)
                self.rates[key] = rate

    def _parse_shapes(self, lines, filename):
        """Parse shape directives"""
        for line in lines:
            if line.startswith("shapes"):
                parts = line.split()
                if len(parts) >= 5:
                    file_path = parts[3]

                    # save absolute path of shapes
                    if not os.path.isabs(file_path):
                        datacard_dir = os.path.dirname(os.path.abspath(filename))
                        file_path = os.path.join(datacard_dir, file_path)

                    if not os.path.exists(file_path):
                        raise FileNotFoundError(
                            f"ROOT file with shapes not found: {file_path}"
                        )

                    shape_info = {
                        "process": parts[1],
                        "channel": parts[2],
                        "file": file_path,
                        "histogram_pattern": parts[4],
                    }
                    self.max_depth = max(self.max_depth, parts[4].count("/"))
                    if len(parts) >= 6:
                        shape_info["histogram_syst_pattern"] = parts[5]
                    self.shapes.append(shape_info)

    def _parse_systematics(self, lines):
        """Parse systematic uncertainty specifications"""
        # Skip lines until after the rate line
        rate_index = None
        for i, line in enumerate(lines):
            if line.startswith("rate"):
                rate_index = i
                break

        if rate_index is None:
            return

        # Parse systematics after the rate line
        for i in range(rate_index + 1, len(lines)):
            line = lines[i]
            logger.debug(f"Now at line {i}: {line}")
            parts = line.split()

            if len(parts) < 2:
                continue

            # Stop when we hit other directives
            if parts[1] in [
                "shapes",
                "nuisance",
                "param",
                "rateParam",
                "group",
                "extArg",
            ]:
                break

            # Check if this looks like a systematic entry (name followed by type)
            if parts[1] in ["lnN", "shape", "gmN", "lnU", "shapeN", "shape?", "shapeU"]:
                syst_name = parts[0]
                syst_type = parts[1]

                syst_info = {"name": syst_name, "type": syst_type, "effects": {}}

                # If gmN, the next value is N
                if syst_type == "gmN" and len(parts) > 2:
                    syst_info["n_events"] = int(parts[2])
                    effects_start = 3
                else:
                    effects_start = 2

                # Get effects on each process
                if len(parts) > effects_start:
                    effects = parts[effects_start:]
                    bin_process_pairs = list(self.rates.keys())

                    if len(effects) == len(bin_process_pairs):
                        for j, effect in enumerate(effects):
                            bin_name, process_name = bin_process_pairs[j]
                            if effect != "-":
                                syst_info["effects"][(bin_name, process_name)] = effect

                self.systematics.append(syst_info)

    def _parse_additional_directives(self, lines):
        """Parse additional directives like param, rateParam, group, etc."""
        for line in lines:
            parts = line.split()
            if len(parts) < 2:
                continue
            reduced_parts = [parts[0], *parts[2:]]
            if parts[1] == "param":
                self.param_lines.append(reduced_parts)
            elif parts[1] == "rateParam":
                self.rate_params.append(reduced_parts)
            elif parts[1] == "group":
                self.group_lines.append(reduced_parts)
            elif parts[1] == "nuisance edit":
                self.nuisance_edits.append(reduced_parts)

    def get_summary(self):
        """Return a summary of the parsed datacard"""
        summary = {
            "channels": self.imax if self.imax else len(self.bins),
            "backgrounds": (
                self.jmax
                if self.jmax
                else len(
                    [p for p in self.processes if self.process_indices.get(p, 0) > 0]
                )
            ),
            "systematics": self.kmax if self.kmax else len(self.systematics),
            "bins": self.bins,
            "observations": self.observations,
            "processes": self.processes,
            "signal_processes": [
                p for p in self.processes if self.process_indices.get(p, 0) <= 0
            ],
            "background_processes": [
                p for p in self.processes if self.process_indices.get(p, 0) > 0
            ],
            "systematics_count": len(self.systematics),
            "has_shapes": len(self.shapes) > 0,
            "rate_params": self.rate_params,
            "param_lines": self.param_lines,
            "group_lines": self.group_lines,
            "nuisance_edits": self.nuisance_edits,
        }
        return summary



import hist
import tensorflow as tf

from rabbit.physicsmodels.physicsmodel import PhysicsModel


class Param(PhysicsModel):
    """
    Custom physics model to transform fit parameters using f(x) = scale*x + offset

    Parameters
    ----------
        indata: Input data used for analysis (e.g., histograms or data structures).
        params: str or list of string
            Parameter names
        scale: float or list of floats, optional
            Scale the parameters
        offset: float or list of floats, optional
            add an offset to the parameters
    """

    need_observables = False
    has_data = False
    has_processes = False
    skip_per_process = True
    skip_prefit = True

    def __init__(
        self,
        indata,
        key,
        params,
        scales=None,
        offsets=None,
    ):
        self.key = key

        param_names = list(indata.signals.astype(str)) + list(indata.systs.astype(str))

        if isinstance(params, str):
            idxs = [param_names.index(params)]
        else:
            idxs = [param_names.index(p) for p in params]

        self.idxs = tf.constant(idxs, dtype=tf.int32)

        self.params = [param_names[i] for i in idxs]

        scales = [scales] if isinstance(int, float) else scales
        offsets = [offsets] if isinstance(int, float) else offsets

        self.scales = tf.constant(scales, tf.float64)
        self.offsets = tf.constant(offsets, tf.float64)

        self.channel_info = {
            "param": {
                "axes": [
                    hist.axis.StrCategory(self.params, name="params"),
                ],
                "start": None,
                "stop": None,
            }
        }

        self.instance = "param"

    @classmethod
    def parse_args(cls, indata, params, scales=None, offsets=None):
        key = " ".join([cls.__name__, params, scales, offsets])
        params = params.split(",")
        if scales is not None:
            scales = [float(s) for s in scales.split(",")]
        if offsets is not None:
            offsets = [float(o) for o in offsets.split(",")]

        return cls(indata, key, params, scales, offsets)

    def compute_flat(self, params):
        params = tf.gather(params, indices=self.idxs, axis=-1)
        return params * self.scales + self.offsets
